import copy
import io
import os
import random
import tempfile

import numpy as np
import onnx
import pytest
import torch

import onnxruntime
import onnxruntime.training.onnxblock as onnxblock
from onnxruntime.capi import _pybind_state as C
from onnxruntime.training import artifacts

# PyTorch Module definitions


def get_opsets_model(filename):
    if isinstance(filename, onnx.ModelProto):
        onx = filename
    else:
        onx = onnx.load(filename)
    return {d.domain: d.version for d in onx.opset_import}


class SimpleNet(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, model_input):
        out = self.fc1(model_input)
        out = self.relu(out)
        out = self.fc2(out)
        return out


# onnxblock Model definitions


class SimpleBlockWithMSELoss(onnxblock.ForwardBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.MSELoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleBlockWithCrossEntropyLoss(onnxblock.ForwardBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.CrossEntropyLoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleTrainingBlockWithMSELoss(onnxblock.TrainingBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.MSELoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleTrainingBlockWithCrossEntropyLoss(onnxblock.TrainingBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.CrossEntropyLoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleBlockWithBCEWithLogitsLoss(onnxblock.ForwardBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.BCEWithLogitsLoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleTrainingBlockWithBCEWithLogitsLoss(onnxblock.TrainingBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.BCEWithLogitsLoss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleBlockWithL1Loss(onnxblock.ForwardBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.L1Loss()

    def build(self, output_name):
        return self.loss(output_name)


class SimpleTrainingBlockWithL1Loss(onnxblock.TrainingBlock):
    def __init__(self):
        super().__init__()
        self.loss = onnxblock.loss.L1Loss()

    def build(self, output_name):
        return self.loss(output_name)


# Test utility methods


def _get_onnx_model(torch_model, model_inputs):
    model_outputs = torch_model(*model_inputs)
    if isinstance(model_outputs, torch.Tensor):
        model_outputs = [model_outputs]
    dynamic_axes = {}
    input_names = []
    output_names = []
    for i, model_input in enumerate(model_inputs):
        input_name = f"input-{i}"
        input_names.append(input_name)
        dynamic_axes[input_name] = {}
        for dim_idx in range(len(model_input.shape)):
            dynamic_axes[input_name].update({dim_idx: f"{input_name}_dim{dim_idx}"})

    for i, model_output in enumerate(model_outputs):
        output_name = f"output-{i}"
        output_names.append(output_name)
        dynamic_axes[output_name] = {}
        for dim_idx in range(len(model_output.shape)):
            dynamic_axes[output_name].update({dim_idx: f"{output_name}_dim{dim_idx}"})

    f = io.BytesIO()
    torch.onnx.export(
        torch_model,
        model_inputs,
        f,
        input_names=input_names,
        output_names=output_names,
        opset_version=14,
        do_constant_folding=False,
        training=torch.onnx.TrainingMode.TRAINING,
        dynamic_axes=dynamic_axes,
        export_params=True,
        keep_initializers_as_inputs=False,
    )
    return onnx.load_model_from_string(f.getvalue())


def _to_numpy(tensor):
    return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()


def _get_models(device, batch_size, input_size, hidden_size, output_size, zero_flag=False):
    """Returns the pt and onnx models for SimpleNet"""
    pt_model = SimpleNet(input_size, hidden_size, output_size).to(device)

    # setting all initial weights to zero
    if zero_flag:
        with torch.no_grad():
            for param in pt_model.parameters():
                param.zero_()

    x = torch.randn(batch_size, input_size, device=device)
    onnx_model = _get_onnx_model(pt_model, (x,))

    return pt_model, onnx_model


def _get_training_ort_output_names(pt_model, onnx_model):
    """Returns the ort output names"""
    ort_output_names = [onnx_model.graph.output[0].name]
    for name, _ in pt_model.named_parameters():
        ort_output_names.append(f"{name}_grad.accumulation.out")

    return ort_output_names


def _get_training_ort_inputs(x, target, pt_model, onnx_model, target_type=None):
    """Returns the ort inputs"""

    ort_inputs = {
        onnx_model.graph.input[0].name: _to_numpy(copy.deepcopy(x)),
        onnx_model.graph.input[1].name: (
            _to_numpy(copy.deepcopy(target))
            if target_type is None
            else _to_numpy(copy.deepcopy(target).type(target_type))
        ),
    }
    if target_type is not None:
        ort_inputs[onnx_model.graph.input[1].name]
    for name, param in pt_model.named_parameters():
        ort_inputs[name] = _to_numpy(copy.deepcopy(param))
        ort_inputs[f"{name}_grad.accumulation.buffer"] = _to_numpy(torch.zeros_like(param))
    ort_inputs["lazy_reset_grad"] = np.full(1, True)

    return ort_inputs


# All unit tests


@pytest.mark.parametrize(
    "block",
    [
        SimpleBlockWithMSELoss,
        SimpleBlockWithCrossEntropyLoss,
        SimpleTrainingBlockWithMSELoss,
        SimpleTrainingBlockWithCrossEntropyLoss,
        SimpleBlockWithBCEWithLogitsLoss,
        SimpleTrainingBlockWithBCEWithLogitsLoss,
        SimpleBlockWithL1Loss,
        SimpleTrainingBlockWithL1Loss,
    ],
)
def test_loss_composition(block):
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    # When / Then no error occurs
    simple_block = block()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)


@pytest.mark.parametrize(
    "block",
    [
        SimpleBlockWithMSELoss,
        SimpleBlockWithL1Loss,
    ],
)
def test_mse_loss_execution(block):
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    # Build the onnx model with loss
    simple_block = block()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model = simple_block.to_model_proto()

    ort_output_names = [onnx_model.graph.output[0].name]
    ort_inputs = {
        onnx_model.graph.input[0].name: _to_numpy(copy.deepcopy(x)),
        onnx_model.graph.input[1].name: _to_numpy(copy.deepcopy(target)),
    }

    def mse_loss(prediction, target):
        loss = torch.nn.MSELoss()
        return loss(prediction, target)

    def l1_loss(prediction, target):
        loss = torch.nn.L1Loss()
        return loss(prediction, target)

    loss = mse_loss if block == SimpleBlockWithMSELoss else l1_loss

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = loss(pt_model(x), target)

        # Then
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


def test_crossentropy_loss_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randint(high=output_size, size=(batch_size,), dtype=torch.int64, device=device)

    # Build the onnx model with loss
    simple_block = SimpleBlockWithCrossEntropyLoss()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model = simple_block.to_model_proto()

    ort_output_names = [onnx_model.graph.output[0].name]
    ort_inputs = {
        onnx_model.graph.input[0].name: _to_numpy(copy.deepcopy(x)),
        onnx_model.graph.input[1].name: _to_numpy(copy.deepcopy(target).type(torch.int64)),
    }

    def crossentropy_loss(prediction, target):
        loss = torch.nn.CrossEntropyLoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = crossentropy_loss(pt_model(x), target)

        # Then
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


def test_bcewithlogits_loss_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    # Build the onnx model with loss
    simple_block = SimpleBlockWithBCEWithLogitsLoss()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model = simple_block.to_model_proto()

    ort_output_names = [onnx_model.graph.output[0].name]
    ort_inputs = {
        onnx_model.graph.input[0].name: _to_numpy(copy.deepcopy(x)),
        onnx_model.graph.input[1].name: _to_numpy(copy.deepcopy(target)),
    }

    def bcewithlogits_loss(prediction, target):
        loss = torch.nn.BCEWithLogitsLoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = bcewithlogits_loss(pt_model(x), target)

        # Then
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


def test_mse_loss_training_graph_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    # Build the onnx trainingmodel with loss
    simple_block = SimpleTrainingBlockWithMSELoss()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model, _ = simple_block.to_model_proto()

    ort_output_names = _get_training_ort_output_names(pt_model, onnx_model)
    ort_inputs = _get_training_ort_inputs(x, target, pt_model, onnx_model)

    def mse_loss(prediction, target):
        loss = torch.nn.MSELoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = mse_loss(pt_model(x), target)
        torch_outs.backward()

        # Then
        # assert loss is close
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


def test_crossentropy_loss_training_graph_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randint(high=output_size, size=(batch_size,), dtype=torch.int64, device=device)

    # Build the onnx trainingmodel with loss
    simple_block = SimpleTrainingBlockWithCrossEntropyLoss()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model, _ = simple_block.to_model_proto()

    ort_output_names = _get_training_ort_output_names(pt_model, onnx_model)
    ort_inputs = _get_training_ort_inputs(x, target, pt_model, onnx_model, target_type=torch.int64)

    def crossentropy_loss(prediction, target):
        loss = torch.nn.CrossEntropyLoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = crossentropy_loss(pt_model(x), target)
        torch_outs.backward()

        # Then
        # assert loss is close
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


def test_bcewithlogits_loss_training_graph_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    # Build the onnx model with loss
    simple_block = SimpleTrainingBlockWithBCEWithLogitsLoss()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model, _ = simple_block.to_model_proto()

    ort_output_names = _get_training_ort_output_names(pt_model, onnx_model)
    ort_inputs = _get_training_ort_inputs(x, target, pt_model, onnx_model)

    def bcewithlogits_loss(prediction, target):
        loss = torch.nn.BCEWithLogitsLoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())

        ort_outs = ort_session.run(ort_output_names, ort_inputs)
        torch_outs = bcewithlogits_loss(pt_model(x), target)
        torch_outs.backward()

        # Then
        # assert loss is close
        assert np.allclose(ort_outs[0], _to_numpy(torch_outs))


@pytest.mark.parametrize(
    "block",
    [SimpleTrainingBlockWithMSELoss, SimpleTrainingBlockWithCrossEntropyLoss, SimpleTrainingBlockWithBCEWithLogitsLoss],
)
@pytest.mark.parametrize("grad_clipping", [None, onnxblock.optim.ClipGradNorm(2.5)])
def test_adamw_optimizer_composition(block, grad_clipping):
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    # When / Then no error occurs
    simple_block = block()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    optimizer = onnxblock.optim.AdamW(clip_grad=grad_clipping)
    with onnxblock.empty_base() as accessor:
        _ = optimizer(simple_block.parameters())
        optimizer_model = accessor.model
        assert optimizer_model


# TODO: Add a test for correctness when creation of ortvalues of
# tensor seq is possible on cuda
def test_adamw_optimizer_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    simple_block = SimpleTrainingBlockWithMSELoss()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    optimizer = onnxblock.optim.AdamW()
    with onnxblock.empty_base() as accessor:
        output_name = optimizer(simple_block.parameters())
        optimizer_model = accessor.model

    learning_rate = 0.001
    step = 1
    ort_output_names = [output_name]

    def mse_loss(prediction, target):
        loss = torch.nn.MSELoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(optimizer_model, onnx_fo.name)

        loss = mse_loss(pt_model(x), target)
        loss.backward()

        ort_inputs = {
            "learning_rate": np.full(1, learning_rate, dtype=np.float32),
            "step": np.full(1, step, dtype=np.int64),
            "params": [],
            "gradients": [],
            "first_order_moments": [],
            "second_order_moments": [],
        }
        for _, param in pt_model.named_parameters():
            ort_inputs["params"].append(_to_numpy(copy.deepcopy(param)))
            ort_inputs["gradients"].append(_to_numpy(copy.deepcopy(param.grad)))
            ort_inputs["first_order_moments"].append(_to_numpy(torch.zeros_like(param)))
            ort_inputs["second_order_moments"].append(_to_numpy(torch.zeros_like(param)))

        # Then no error occurs when executing the model
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())
        _ = ort_session.run(ort_output_names, ort_inputs)


@pytest.mark.parametrize(
    "block",
    [SimpleTrainingBlockWithMSELoss, SimpleTrainingBlockWithCrossEntropyLoss, SimpleTrainingBlockWithBCEWithLogitsLoss],
)
@pytest.mark.parametrize("grad_clipping", [None, onnxblock.optim.ClipGradNorm(2.5)])
def test_sgd_optimizer_composition(block, grad_clipping):
    # Given
    device = "cpu"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    # When / Then no error occurs
    simple_block = block()
    for name, _ in pt_model.named_parameters():
        simple_block.requires_grad(name)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    optimizer = onnxblock.optim.SGD(clip_grad=grad_clipping)
    with onnxblock.empty_base() as accessor:
        _ = optimizer(simple_block.parameters())
        optimizer_model = accessor.model
        assert optimizer_model


def test_retrieve_parameters():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    simple_block = SimpleTrainingBlockWithMSELoss()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    # When
    trainable_params, non_trainable_params = simple_block.parameters()

    # Then
    assert not non_trainable_params
    for ort_param, (pt_param_name, pt_param) in zip(trainable_params, pt_model.named_parameters(), strict=False):
        assert ort_param.name == pt_param_name
        assert np.allclose(
            np.frombuffer(ort_param.raw_data, dtype=np.float32).reshape(pt_param.shape),
            _to_numpy(pt_param),
        )


def test_retrieve_parameters_before_building_gradient_graph():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    simple_block = SimpleTrainingBlockWithMSELoss()

    # When / Then
    with pytest.raises(Exception) as ex_info:
        _, _ = simple_block.parameters()
    assert "Please build the training model first before trying to retrieve the parameters." in str(ex_info.value)


def test_save_checkpoint():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    simple_block = SimpleTrainingBlockWithMSELoss()
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)
    trainable_params, non_trainable_params = simple_block.parameters()

    # When
    with tempfile.TemporaryDirectory() as checkpoint_dir_name:
        checkpoint_file_path = os.path.join(checkpoint_dir_name, "checkpoint")
        onnxblock.save_checkpoint((trainable_params, non_trainable_params), checkpoint_file_path)
        # Then
        assert os.path.exists(checkpoint_file_path)


def test_load_checkpoint():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, zero_onnx_model = _get_models(device, batch_size, input_size, hidden_size, output_size, zero_flag=True)
    for initializer in zero_onnx_model.graph.initializer:
        zero_np = onnx.numpy_helper.to_array(initializer)
        assert np.allclose(zero_np, np.zeros(zero_np.shape))

    _, onnx_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    # Copy of onnx_model for comparison
    onnx_model_copy = copy.deepcopy(onnx_model)

    simple_block = SimpleTrainingBlockWithMSELoss()

    # When
    simple_block.requires_grad("fc2.weight", True)
    simple_block.requires_grad("fc1.bias", True)
    simple_block.requires_grad("fc1.weight", True)
    simple_block.requires_grad("fc2.bias", True)

    with onnxblock.base(onnx_model):
        _ = simple_block(onnx_model.graph.output[0].name)
    trainable_params, non_trainable_params = simple_block.parameters()

    with tempfile.TemporaryDirectory() as checkpoint_dir_name:
        checkpoint_file_path = os.path.join(checkpoint_dir_name, "checkpoint")
        onnxblock.save_checkpoint((trainable_params, non_trainable_params), checkpoint_file_path)

        # Load checkpoint parameters to the new simple model
        onnxblock.load_checkpoint_to_model(checkpoint_file_path, zero_onnx_model)

        # Then
        onnx_model_copy.graph.initializer.sort(key=lambda x: x.name)
        zero_onnx_model.graph.initializer.sort(key=lambda x: x.name)

        for i, initializer in enumerate(onnx_model_copy.graph.initializer):
            if initializer.name in ["fc2.weight", "fc1.bias"]:
                onnx_np = onnx.numpy_helper.to_array(onnx_model_copy.graph.initializer[i])
                zero_np = onnx.numpy_helper.to_array(zero_onnx_model.graph.initializer[i])
                assert np.allclose(onnx_np, zero_np)


def test_set_requires_grad_on_parameters():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    simple_block = SimpleTrainingBlockWithMSELoss()

    # When
    simple_block.requires_grad("fc2.weight", False)
    simple_block.requires_grad("fc1.bias", False)

    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)
    trainable_params, non_trainable_params = simple_block.parameters()

    # Then
    expected_trainable_parameters = {"fc1.weight", "fc2.bias"}
    expected_non_trainable_parameters = {"fc2.weight", "fc1.bias"}
    for param in trainable_params:
        assert param.name in expected_trainable_parameters
    for param in non_trainable_params:
        assert param.name in expected_non_trainable_parameters


def test_set_requires_grad_on_inputs():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    # When
    simple_block = SimpleTrainingBlockWithMSELoss()
    simple_block.requires_grad("input-0")
    with onnxblock.base(base_model):
        _ = simple_block(base_model.graph.output[0].name)

    onnx_model, _ = simple_block.to_model_proto()

    # Then
    expected_gradient_buffer_name = "input-0_grad.accumulation.buffer"
    expectedt_gradient_output_name = "input-0_grad.accumulation.out"
    graph_input_names = {graph_input.name for graph_input in onnx_model.graph.input}
    graph_output_names = {graph_output.name for graph_output in onnx_model.graph.output}

    assert expected_gradient_buffer_name in graph_input_names
    assert expectedt_gradient_output_name in graph_output_names


@pytest.mark.parametrize("model_type", [onnxblock.ForwardBlock, onnxblock.TrainingBlock])
def test_weighted_average_model_composition(model_type):
    # Given
    class TwoOutputNet(torch.nn.Module):
        def __init__(self, input_size, hidden_size, num_classes):
            super().__init__()

            self.fc1_1 = torch.nn.Linear(input_size, hidden_size)
            self.relu1 = torch.nn.ReLU()
            self.fc1_2 = torch.nn.Linear(hidden_size, num_classes)

            self.fc2_1 = torch.nn.Linear(input_size, hidden_size)
            self.relu2 = torch.nn.ReLU()
            self.fc2_2 = torch.nn.Linear(hidden_size, num_classes)

        def forward(self, model_input1, model_input2):
            out1 = self.fc1_2(self.relu1(self.fc1_1(model_input1)))
            out2 = self.fc2_2(self.relu2(self.fc2_1(model_input2)))
            return out1, out2

    class WeightedAvg(model_type):
        def __init__(self, w1, w2):
            super().__init__()

            self.loss1 = onnxblock.loss.CrossEntropyLoss()
            self.loss2 = onnxblock.loss.CrossEntropyLoss()
            self.w1 = onnxblock.blocks.Constant(w1)
            self.w2 = onnxblock.blocks.Constant(w2)
            self.mul = onnxblock.blocks.Mul()
            self.add = onnxblock.blocks.Add()

        def build(self, loss_input_name1, loss_input_name2):
            return self.add(
                self.mul(self.w1(), self.loss1(loss_input_name1, labels_name="labels1")),
                self.mul(self.w2(), self.loss2(loss_input_name2, labels_name="labels2")),
            )

    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model = TwoOutputNet(input_size, hidden_size, output_size).to(device)
    x1 = torch.randn(batch_size, input_size, device=device)
    x2 = torch.randn(batch_size, input_size, device=device)
    base_model = _get_onnx_model(pt_model, (x1, x2))

    # When / Then no error occurs
    weighted_model = WeightedAvg(random.random(), random.random())
    with onnxblock.base(base_model):
        _ = weighted_model(base_model.graph.output[0].name, base_model.graph.output[1].name)


def test_grad_clipping_execution():
    # Given
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    pt_model, _ = _get_models(device, batch_size, input_size, hidden_size, output_size)
    x = torch.randn(batch_size, input_size, device=device)
    target = torch.randn(batch_size, output_size, device=device)

    # Prepare the onnx model with only grad clipping
    base_model = onnx.ModelProto()
    base_model.graph.name = "ClipGradNorm Model"
    base_model.producer_name = "grad clipping test"
    base_model.opset_import.extend(
        [
            onnx.helper.make_opsetid("com.microsoft", 1),
            onnx.helper.make_opsetid("", onnx.defs.onnx_opset_version()),
        ]
    )
    base_model.ir_version = onnx.IR_VERSION

    class GradClippingModel(onnxblock.ForwardBlock):
        def __init__(self, max_norm):
            super().__init__()
            self._grad_clip = onnxblock.optim.ClipGradNorm(max_norm)

        def build(self, grads_name):
            return self._grad_clip(grads_name)

    base_model.graph.input.append(
        onnx.helper.make_tensor_sequence_value_info("gradients", onnx.TensorProto.FLOAT, None)
    )

    grad_clip = GradClippingModel(2.5)
    with onnxblock.base(base_model):
        ort_output_names = grad_clip("gradients")

    onnx_model = grad_clip.to_model_proto()

    onnx_model.graph.output.append(
        onnx.helper.make_tensor_sequence_value_info(ort_output_names, onnx.TensorProto.FLOAT, None)
    )

    def mse_loss(prediction, target):
        loss = torch.nn.MSELoss()
        return loss(prediction, target)

    # When
    with tempfile.NamedTemporaryFile(suffix=".onnx") as onnx_fo:
        onnx.save(onnx_model, onnx_fo.name)

        loss = mse_loss(pt_model(x), target)
        loss.backward()

        ort_inputs = {"gradients": []}
        for _, param in pt_model.named_parameters():
            ort_inputs["gradients"].append(_to_numpy(copy.deepcopy(param.grad)))

        torch.nn.utils.clip_grad_norm_(pt_model.parameters(), 2.5)

        # Then no error occurs when executing the model
        ort_session = onnxruntime.InferenceSession(onnx_fo.name, providers=C.get_available_providers())
        ort_outs = ort_session.run([ort_output_names], ort_inputs)

        # assert all the gradients are close
        for ort_grad, pt_param in zip(ort_outs[0], pt_model.parameters(), strict=False):
            assert np.allclose(ort_grad, _to_numpy(pt_param.grad))


def test_additional_output_names():
    class DropoutModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.dropout = torch.nn.Dropout(p=0.5)

        def forward(self, x):
            return self.dropout(x)

    model = DropoutModel()
    onnx_model = _get_onnx_model(model, (torch.randn(1, 3, 224, 224),))

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(onnx_model, loss=artifacts.LossType.CrossEntropyLoss, artifact_directory=temp_dir)

        eval_model = onnx.load(os.path.join(temp_dir, "eval_model.onnx"))

        # Make sure only loss is the output
        assert len(eval_model.graph.output) == 1

        # Re-generate artifacts with additional output names
        artifacts.generate_artifacts(
            onnx_model,
            loss=artifacts.LossType.CrossEntropyLoss,
            artifact_directory=temp_dir,
            additional_output_names=["output-0"],
        )

        # Make sure the eval model has two outputs
        eval_model = onnx.load(os.path.join(temp_dir, "eval_model.onnx"))
        assert len(eval_model.graph.output) == 2


def test_eval_model_has_no_training_mode_dropout():
    class DropoutModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.dropout = torch.nn.Dropout(p=0.5)

        def forward(self, x):
            return self.dropout(x)

    model = DropoutModel()
    onnx_model = _get_onnx_model(model, (torch.randn(1, 3, 224, 224),))

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(onnx_model, loss=artifacts.LossType.CrossEntropyLoss, artifact_directory=temp_dir)

        eval_model = onnx.load(os.path.join(temp_dir, "eval_model.onnx"))

        flag = False
        for node in eval_model.graph.node:
            if node.op_type == "Dropout":
                assert not node.input[2]
                flag = True

        assert flag


def test_eval_model_has_no_training_mode_batchnorm():
    class BatchNormModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.batchnorm = torch.nn.BatchNorm2d(100)

        def forward(self, x):
            return self.batchnorm(x)

    model = BatchNormModel()
    onnx_model = _get_onnx_model(model, (torch.randn(20, 100, 35, 45),))

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(onnx_model, loss=artifacts.LossType.CrossEntropyLoss, artifact_directory=temp_dir)

        eval_model = onnx.load(os.path.join(temp_dir, "eval_model.onnx"))

        flag = False
        for node in eval_model.graph.node:
            if node.op_type == "BatchNormalization":
                for attr in node.attribute:
                    if attr.name == "training_mode":
                        assert attr.i == 0
                        flag = True

        assert flag


def test_label_encoder_composition():
    device = "cuda"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)
    base_model.opset_import.append(
        onnx.helper.make_opsetid("ai.onnx.ml", onnx.defs.onnx_opset_version()),
    )

    all_nodes = [node.op_type for node in base_model.graph.node]
    assert "LabelEncoder" not in all_nodes

    class SCELossWithLabelEncoder(onnxblock.ForwardBlock):
        def __init__(self):
            super().__init__()
            self._loss = onnxblock.loss.CrossEntropyLoss()

        def build(self, output_name):
            keys_int64s = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            values_int64s = [521, 522, 523, 524, 525, 526, 527, 528, 529, 530]

            # Create a new graph input for the labels
            labels_name = "labels"
            labels_input = copy.deepcopy(self.base.graph.output[0])
            labels_input.type.tensor_type.elem_type = onnx.TensorProto.INT64
            labels_input.name = labels_name
            del labels_input.type.tensor_type.shape.dim[1]
            self.base.graph.input.append(labels_input)

            label_encoder = onnxblock.blocks.LabelEncoder(
                default_int64=521, keys_int64s=keys_int64s, values_int64s=values_int64s
            )

            return self._loss(output_name, label_encoder(labels_name))

    block = SCELossWithLabelEncoder()
    model = None
    with onnxblock.base(base_model):
        _ = block(base_model.graph.output[0].name)
        model = block.to_model_proto()

    all_nodes = [node.op_type for node in model.graph.node]
    assert "LabelEncoder" in all_nodes


def test_save_ort_format():
    device = "cpu"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(
            base_model,
            requires_grad=["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"],
            loss=artifacts.LossType.CrossEntropyLoss,
            optimizer=artifacts.OptimType.AdamW,
            artifact_directory=temp_dir,
            ort_format=True,
        )

        assert os.path.exists(os.path.join(temp_dir, "training_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "training_model.ort"))
        assert os.path.exists(os.path.join(temp_dir, "eval_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "eval_model.ort"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.ort"))
        base_opsets = get_opsets_model(base_model)
        training_opsets = get_opsets_model(os.path.join(temp_dir, "training_model.onnx"))
        eval_opsets = get_opsets_model(os.path.join(temp_dir, "eval_model.onnx"))
        optimizer_opsets = get_opsets_model(os.path.join(temp_dir, "optimizer_model.onnx"))
        if base_opsets[""] != training_opsets[""]:
            raise AssertionError(f"Opsets mismatch {base_opsets['']} != {training_opsets['']}.")
        if base_opsets[""] != eval_opsets[""]:
            raise AssertionError(f"Opsets mismatch {base_opsets['']} != {eval_opsets['']}.")
        if base_opsets[""] != optimizer_opsets[""]:
            raise AssertionError(f"Opsets mismatch {base_opsets['']} != {optimizer_opsets['']}.")


def test_custom_loss_function():
    # This test tries to add a custom loss function to the model.
    # The custom loss function tries to use two model outputs of two different ranks, computes the
    # two losses and returns the sum of the two losses.
    # If the artifacts are generated successfully, without an exception being raised, the test passes.
    class ModelWithTwoOutputs(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.a = torch.randn(20, 100, 35, 45)
            self.b = torch.randn(40, 100, 70)

        def forward(self, x, y):
            return self.a + x, self.b + y

    class CustomLossBlock(onnxblock.Block):
        def __init__(self):
            self._loss1 = onnxblock.loss.MSELoss()
            self._loss2 = onnxblock.loss.BCEWithLogitsLoss()
            self._add = onnxblock.blocks.Add()

        def build(self, input1, input2):
            return self._add(self._loss1(input1, target_name="target1"), self._loss2(input2, target_name="target2"))

    model = ModelWithTwoOutputs()
    onnx_model = _get_onnx_model(model, (torch.randn(20, 100, 35, 45), torch.randn(40, 100, 70)))

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(onnx_model, loss=CustomLossBlock(), artifact_directory=temp_dir)


def test_save_nominal_checkpoint():
    device = "cpu"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(
            base_model,
            requires_grad=["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"],
            loss=artifacts.LossType.CrossEntropyLoss,
            optimizer=artifacts.OptimType.AdamW,
            artifact_directory=temp_dir,
            nominal_checkpoint=True,
        )

        assert os.path.exists(os.path.join(temp_dir, "checkpoint"))
        assert os.path.exists(os.path.join(temp_dir, "nominal_checkpoint"))
        assert (
            os.stat(os.path.join(temp_dir, "checkpoint")).st_size
            > os.stat(os.path.join(temp_dir, "nominal_checkpoint")).st_size
        )


def test_custom_optimizer_block():
    device = "cpu"
    batch_size, input_size, hidden_size, output_size = 64, 784, 500, 10
    _, base_model = _get_models(device, batch_size, input_size, hidden_size, output_size)
    weight_decay = 123
    optimizer = onnxblock.optim.AdamW(weight_decay=weight_decay)

    with tempfile.TemporaryDirectory() as temp_dir:
        artifacts.generate_artifacts(
            base_model,
            requires_grad=["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"],
            loss=artifacts.LossType.CrossEntropyLoss,
            optimizer=optimizer,
            artifact_directory=temp_dir,
        )

        assert os.path.exists(os.path.join(temp_dir, "checkpoint"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.onnx"))

        optimizer_model = onnx.load(os.path.join(temp_dir, "optimizer_model.onnx"))
        for node in optimizer_model.graph.node:
            if node.op_type == "AdamW":
                for attr in node.attribute:
                    if attr.name == "weight_decay":
                        assert attr.f == weight_decay


def test_generate_artifacts_path():
    with tempfile.TemporaryDirectory() as temp_dir:
        _, simple_net = _get_models("cpu", 32, 28, 10, 10)

        requires_grad_params = ["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"]

        onnx.save_model(
            simple_net,
            os.path.join(temp_dir, "simple_net.onnx"),
        )

        artifacts.generate_artifacts(
            os.path.join(temp_dir, "simple_net.onnx"),
            requires_grad=requires_grad_params,
            loss=artifacts.LossType.CrossEntropyLoss,
            optimizer=artifacts.OptimType.AdamW,
            artifact_directory=temp_dir,
        )

        # generate_artifacts should have thrown if it didn't complete successfully.
        # Below is a sanity check to validate that all the expected files were created.
        assert os.path.exists(os.path.join(temp_dir, "training_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "eval_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "checkpoint"))


def test_generate_artifacts_external_data_one_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        _, simple_net = _get_models("cpu", 32, 28, 10, 10)

        requires_grad_params = ["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"]

        onnx.save_model(
            simple_net,
            os.path.join(temp_dir, "simple_net.onnx"),
            save_as_external_data=True,
            all_tensors_to_one_file=True,
            size_threshold=0,
        )

        artifacts.generate_artifacts(
            os.path.join(temp_dir, "simple_net.onnx"),
            requires_grad=requires_grad_params,
            loss=artifacts.LossType.CrossEntropyLoss,
            optimizer=artifacts.OptimType.AdamW,
            artifact_directory=temp_dir,
        )

        # generate_artifacts should have thrown if it didn't complete successfully.
        # Below is a sanity check to validate that all the expected files were created.
        assert os.path.exists(os.path.join(temp_dir, "training_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "eval_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "checkpoint"))


@pytest.mark.parametrize("loss", list(artifacts.LossType))
def test_generate_artifacts_external_data_separate_files(loss):
    with tempfile.TemporaryDirectory() as temp_dir:
        _, simple_net = _get_models("cpu", 32, 28, 10, 10)

        requires_grad_params = ["fc1.weight", "fc1.bias", "fc2.weight", "fc2.bias"]

        onnx.save_model(
            simple_net,
            os.path.join(temp_dir, "simple_net.onnx"),
            save_as_external_data=True,
            all_tensors_to_one_file=False,
            size_threshold=0,
        )

        artifacts.generate_artifacts(
            os.path.join(temp_dir, "simple_net.onnx"),
            requires_grad=requires_grad_params,
            loss=loss,
            optimizer=artifacts.OptimType.AdamW,
            artifact_directory=temp_dir,
        )

        # generate_artifacts should have thrown if it didn't complete successfully.
        # Below is a sanity check to validate that all the expected files were created.
        assert os.path.exists(os.path.join(temp_dir, "training_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "eval_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "optimizer_model.onnx"))
        assert os.path.exists(os.path.join(temp_dir, "checkpoint"))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import unittest

import onnx
from conformer_model_generator import create_conformer_attention
from parity_utilities import find_transformers_source

if find_transformers_source():
    from fusion_options import FusionOptions
    from onnx_model import OnnxModel
    from optimizer import optimize_model
else:
    from onnxruntime.transformers.fusion_options import FusionOptions
    from onnxruntime.transformers.onnx_model import OnnxModel
    from onnxruntime.transformers.optimizer import optimize_model


class TestFusion(unittest.TestCase):
    def verify_fusion(self, optimized_model, expected_model_filename):
        optimized_model.topological_sort(is_deterministic=True)

        expected_model_path = os.path.join(
            os.path.dirname(__file__), "test_data", "models", "conformer", expected_model_filename
        )
        print("Expected model path = ", expected_model_path)
        expected_model = OnnxModel(onnx.load(expected_model_path))
        expected_model.topological_sort(is_deterministic=True)

        nodes = optimized_model.model.graph.node
        self.assertEqual(len(nodes), len(expected_model.model.graph.node))

        for i in range(len(nodes)):
            self.assertEqual(nodes[i], expected_model.model.graph.node[i])

        for expected_initializer in expected_model.model.graph.initializer:
            print("Expected initializer initial = ", expected_initializer.name)
            self.assertTrue(
                OnnxModel.has_same_value(
                    optimized_model.get_initializer(expected_initializer.name), expected_initializer
                )
            )

    def test_ct_mha_fusion(self):
        num_heads = 8
        hidden_size = 512
        model = create_conformer_attention(num_heads=num_heads, hidden_size=hidden_size, add_before_layernorm=False)
        dir = "."
        model_path = os.path.join(dir, "conformer_self_mha.onnx")
        onnx.save(model, model_path)
        options = FusionOptions("conformer")
        optimized_model = optimize_model(
            model_path,
            model_type="conformer",
            num_heads=num_heads,
            hidden_size=hidden_size,
            optimization_options=options,
        )
        os.remove(model_path)
        self.verify_fusion(optimized_model, "conformer_self_mha_fused.onnx")


if __name__ == "__main__":
    unittest.main()



"""
*Experimental* implementation of FlashAttention in Triton.
Tested with triton==2.0.0.dev20221202.
Triton 2.0 has a new backend (MLIR) but seems like it doesn't yet work for head dimensions
other than 64:
https://github.com/openai/triton/blob/d376020f90002757eea3ea9475d4f7cfc2ec5ead/python/triton/ops/flash_attention.py#L207
We'll update this implementation with the new Triton backend once this is fixed.

We use the FlashAttention implementation from Phil Tillet a starting point.
https://github.com/openai/triton/blob/master/python/tutorials/06-fused-attention.py

Changes:
- Implement both causal and non-causal attention.
- Implement both self-attention and cross-attention.
- Support arbitrary seqlens (not just multiples of 128), for both forward and backward.
- Support all head dimensions up to 128 (not just 16, 32, 64, 128), for both forward and backward.
- Support attention bias.
- Speed up the forward pass a bit, and only store the LSE instead of m and l.
- Make the backward for d=128 much faster by reducing register spilling.
- Optionally parallelize the backward pass across seqlen_k, to deal with the case of
small batch size * nheads.

Caution:
- This is an *experimental* implementation. The forward pass should be quite robust but
I'm not 100% sure that the backward pass doesn't have race conditions (due to the Triton compiler).
- This implementation has only been tested on A100.
- If you plan to use headdim other than 64 and 128, you should test for race conditions
(due to the Triton compiler), as done in tests/test_flash_attn.py
"test_flash_attn_triton_race_condition". I've tested and fixed many race conditions
for different head dimensions (40, 48, 64, 128, 80, 88, 96), but I'm still not 100% confident
that there are none left for other head dimensions.

Differences between this Triton version and the CUDA version:
- Triton version doesn't support dropout.
- Triton forward is generally faster than CUDA forward, while Triton backward is
generally slower than CUDA backward. Overall Triton forward + backward is slightly slower
than CUDA forward + backward.
- Triton version doesn't support different sequence lengths in a batch (i.e., RaggedTensor/NestedTensor).
- Triton version supports attention bias, while CUDA version doesn't.
"""

import math

import torch
import triton
import triton.language as tl
from onnx import GraphProto, NodeProto, TensorProto, helper

from onnxruntime.training.ortmodule import register_graph_optimizer
from onnxruntime.training.ortmodule.graph_optimizers.utils import GraphMatcher, check_attribute_value, update_graph


# Disabling autotune for now, set num_warps=4 if headdim=64 and num_warps=8 if headdim=128
# @triton.autotune(
#     configs=[
#         triton.Config({"BLOCK_M": 128, "BLOCK_N": 128}, num_warps=4, num_stages=1),
#         # This config has a race condition when EVEN_M == False, disabling it for now.
#         # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64}, num_warps=4, num_stages=1),
#     ],
#     key=['CACHE_KEY_SEQLEN_Q', 'CACHE_KEY_SEQLEN_K', 'BIAS_TYPE', 'IS_CAUSAL', 'BLOCK_HEADDIM']
# )
@triton.heuristics(
    {
        "EVEN_M": lambda args: args["seqlen_q"] % args["BLOCK_M"] == 0,
        "EVEN_N": lambda args: args["seqlen_k"] % args["BLOCK_N"] == 0,
        "EVEN_HEADDIM": lambda args: args["headdim"] == args["BLOCK_HEADDIM"],
    }
)
@triton.jit
def _fwd_kernel(
    Q,
    K,
    V,
    Bias,
    Out,
    Lse,
    TMP,  # NOTE: TMP is a scratchpad buffer to workaround a compiler bug
    softmax_scale,
    stride_qb,
    stride_qh,
    stride_qm,
    stride_kb,
    stride_kh,
    stride_kn,
    stride_vb,
    stride_vh,
    stride_vn,
    stride_bb,
    stride_bh,
    stride_bm,
    stride_ob,
    stride_oh,
    stride_om,
    nheads,
    seqlen_q,
    seqlen_k,
    seqlen_q_rounded,
    headdim,
    CACHE_KEY_SEQLEN_Q,
    CACHE_KEY_SEQLEN_K,
    BIAS_TYPE: tl.constexpr,
    IS_CAUSAL: tl.constexpr,
    BLOCK_HEADDIM: tl.constexpr,
    EVEN_M: tl.constexpr,
    EVEN_N: tl.constexpr,
    EVEN_HEADDIM: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    start_m = tl.program_id(0)
    off_hb = tl.program_id(1)
    off_b = off_hb // nheads
    off_h = off_hb % nheads
    # initialize offsets
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_n = tl.arange(0, BLOCK_N)
    offs_d = tl.arange(0, BLOCK_HEADDIM)
    # Initialize pointers to Q, K, V
    # Adding parenthesis around indexing might use int32 math instead of int64 math?
    # https://github.com/openai/triton/issues/741
    # I'm seeing a tiny bit of difference (5-7us)
    q_ptrs = Q + off_b * stride_qb + off_h * stride_qh + (offs_m[:, None] * stride_qm + offs_d[None, :])
    k_ptrs = K + off_b * stride_kb + off_h * stride_kh + (offs_n[:, None] * stride_kn + offs_d[None, :])
    v_ptrs = V + off_b * stride_vb + off_h * stride_vh + (offs_n[:, None] * stride_vn + offs_d[None, :])
    if BIAS_TYPE == "vector":
        b_ptrs = Bias + off_b * stride_bb + off_h * stride_bh + offs_n
    elif BIAS_TYPE == "matrix":
        b_ptrs = Bias + off_b * stride_bb + off_h * stride_bh + (offs_m[:, None] * stride_bm + offs_n[None, :])
    # initialize pointer to m and l
    t_ptrs = TMP + off_hb * seqlen_q_rounded + offs_m
    lse_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    acc_o = tl.zeros([BLOCK_M, BLOCK_HEADDIM], dtype=tl.float32)
    # load q: it will stay in SRAM throughout
    # [2022-10-30] TD: Triton bug - in the case of EVEN_M=True and EVEN_N=False, if we just call
    # tl.load(q_ptrs), we get the wrong output!
    if EVEN_M & EVEN_N:
        if EVEN_HEADDIM:
            q = tl.load(q_ptrs)
        else:
            q = tl.load(q_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
    else:
        if EVEN_HEADDIM:
            q = tl.load(q_ptrs, mask=offs_m[:, None] < seqlen_q, other=0.0)
        else:
            q = tl.load(q_ptrs, mask=(offs_m[:, None] < seqlen_q) & (offs_d[None, :] < headdim), other=0.0)
    # loop over k, v and update accumulator
    end_n = seqlen_k if not IS_CAUSAL else tl.minimum((start_m + 1) * BLOCK_M, seqlen_k)
    for start_n in range(0, end_n, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)
        # -- compute qk ----
        if EVEN_N & EVEN_M:  # If we just do "if EVEN_N", there seems to be some race condition
            if EVEN_HEADDIM:
                k = tl.load(k_ptrs + start_n * stride_kn)
            else:
                k = tl.load(k_ptrs + start_n * stride_kn, mask=offs_d[None, :] < headdim, other=0.0)
        else:
            if EVEN_HEADDIM:
                k = tl.load(
                    k_ptrs + start_n * stride_kn,
                    mask=(start_n + offs_n)[:, None] < seqlen_k,
                    other=0.0,
                )
            else:
                k = tl.load(
                    k_ptrs + start_n * stride_kn,
                    mask=((start_n + offs_n)[:, None] < seqlen_k) & (offs_d[None, :] < headdim),
                    other=0.0,
                )
        qk = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        qk += tl.dot(q, k, trans_b=True)
        # Trying to combine the two masks seem to make the result wrong
        if not EVEN_N:  # Need to mask out otherwise the softmax is wrong
            qk += tl.where((start_n + offs_n)[None, :] < seqlen_k, 0, float("-inf"))
        if IS_CAUSAL:
            qk += tl.where(offs_m[:, None] >= (start_n + offs_n)[None, :], 0, float("-inf"))
        if BIAS_TYPE != "none":
            if BIAS_TYPE == "vector":
                if EVEN_N:
                    bias = tl.load(b_ptrs + start_n).to(tl.float32)
                else:
                    bias = tl.load(b_ptrs + start_n, mask=(start_n + offs_n) < seqlen_k, other=0.0).to(tl.float32)
                bias = bias[None, :]
            elif BIAS_TYPE == "matrix":
                if EVEN_M & EVEN_N:
                    bias = tl.load(b_ptrs + start_n).to(tl.float32)
                else:
                    bias = tl.load(
                        b_ptrs + start_n,
                        mask=(offs_m[:, None] < seqlen_q) & ((start_n + offs_n)[None, :] < seqlen_k),
                        other=0.0,
                    ).to(tl.float32)
            # Slightly faster to multiply the softmax_scale in the tl.exp below since the compiler
            # can then fuse the mult and add into an fma instruction. But if we have bias we need to
            # to multiply with softmax_scale here.
            qk = qk * softmax_scale + bias
            m_ij = tl.maximum(tl.max(qk, 1), lse_i)
            p = tl.exp(qk - m_ij[:, None])
        else:
            m_ij = tl.maximum(tl.max(qk, 1) * softmax_scale, lse_i)
            p = tl.exp(qk * softmax_scale - m_ij[:, None])
        l_ij = tl.sum(p, 1)

        # scale acc_o
        acc_o_scale = tl.exp(m_i - m_ij)

        # # -- update output accumulator --
        # BUG: have to store and immediately load
        tl.store(t_ptrs, acc_o_scale)
        acc_o_scale = tl.load(t_ptrs)
        acc_o = acc_o * acc_o_scale[:, None]
        # update acc_o
        if EVEN_N & EVEN_M:  # If we just do "if EVEN_N", there seems to be some race condition
            if EVEN_HEADDIM:
                v = tl.load(v_ptrs + start_n * stride_vn)
            else:
                v = tl.load(v_ptrs + start_n * stride_vn, mask=offs_d[None, :] < headdim, other=0.0)
        else:
            if EVEN_HEADDIM:
                v = tl.load(
                    v_ptrs + start_n * stride_vn,
                    mask=(start_n + offs_n)[:, None] < seqlen_k,
                    other=0.0,
                )
            else:
                v = tl.load(
                    v_ptrs + start_n * stride_vn,
                    mask=((start_n + offs_n)[:, None] < seqlen_k) & (offs_d[None, :] < headdim),
                    other=0.0,
                )
        p = p.to(v.dtype)
        acc_o += tl.dot(p, v)

        # -- update statistics
        m_i = m_ij
        l_i_new = tl.exp(lse_i - m_ij) + l_ij
        lse_i = m_ij + tl.log(l_i_new)

    o_scale = tl.exp(m_i - lse_i)
    # BUG: have to store and immediately load
    tl.store(t_ptrs, o_scale)
    o_scale = tl.load(t_ptrs)
    acc_o = acc_o * o_scale[:, None]
    # rematerialize offsets to save registers
    start_m = tl.program_id(0)
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    # write back l and m
    lse_ptrs = Lse + off_hb * seqlen_q_rounded + offs_m
    tl.store(lse_ptrs, lse_i)
    # initialize pointers to output
    offs_d = tl.arange(0, BLOCK_HEADDIM)
    out_ptrs = Out + off_b * stride_ob + off_h * stride_oh + (offs_m[:, None] * stride_om + offs_d[None, :])
    if EVEN_M:
        if EVEN_HEADDIM:
            tl.store(out_ptrs, acc_o)
        else:
            tl.store(out_ptrs, acc_o, mask=offs_d[None, :] < headdim)
    else:
        if EVEN_HEADDIM:
            tl.store(out_ptrs, acc_o, mask=offs_m[:, None] < seqlen_q)
        else:
            tl.store(out_ptrs, acc_o, mask=(offs_m[:, None] < seqlen_q) & (offs_d[None, :] < headdim))


@triton.jit
def _bwd_preprocess_do_o_dot(
    Out,
    DO,
    Delta,
    stride_ob,
    stride_oh,
    stride_om,
    stride_dob,
    stride_doh,
    stride_dom,
    nheads,
    seqlen_q,
    seqlen_q_rounded,
    headdim,
    BLOCK_M: tl.constexpr,
    BLOCK_HEADDIM: tl.constexpr,
):
    start_m = tl.program_id(0)
    off_hb = tl.program_id(1)
    off_b = off_hb // nheads
    off_h = off_hb % nheads
    # initialize offsets
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_d = tl.arange(0, BLOCK_HEADDIM)
    # load
    o = tl.load(
        Out + off_b * stride_ob + off_h * stride_oh + offs_m[:, None] * stride_om + offs_d[None, :],
        mask=(offs_m[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
        other=0.0,
    ).to(tl.float32)
    do = tl.load(
        DO + off_b * stride_dob + off_h * stride_doh + offs_m[:, None] * stride_dom + offs_d[None, :],
        mask=(offs_m[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
        other=0.0,
    ).to(tl.float32)
    delta = tl.sum(o * do, axis=1)
    # write-back
    tl.store(Delta + off_hb * seqlen_q_rounded + offs_m, delta)


@triton.jit
def _bwd_store_dk_dv(
    dk_ptrs,
    dv_ptrs,
    dk,
    dv,
    offs_n,
    offs_d,
    seqlen_k,
    headdim,
    EVEN_M: tl.constexpr,
    EVEN_N: tl.constexpr,
    EVEN_HEADDIM: tl.constexpr,
):
    # [2022-11-01] TD: Same bug. In the case of EVEN_N=True and EVEN_M=False,
    # if we just call tl.store(dv_ptrs), there's a race condition
    if EVEN_N & EVEN_M:
        if EVEN_HEADDIM:
            tl.store(dv_ptrs, dv)
            tl.store(dk_ptrs, dk)
        else:
            tl.store(dv_ptrs, dv, mask=offs_d[None, :] < headdim)
            tl.store(dk_ptrs, dk, mask=offs_d[None, :] < headdim)
    else:
        if EVEN_HEADDIM:
            tl.store(dv_ptrs, dv, mask=offs_n[:, None] < seqlen_k)
            tl.store(dk_ptrs, dk, mask=offs_n[:, None] < seqlen_k)
        else:
            tl.store(dv_ptrs, dv, mask=(offs_n[:, None] < seqlen_k) & (offs_d[None, :] < headdim))
            tl.store(dk_ptrs, dk, mask=(offs_n[:, None] < seqlen_k) & (offs_d[None, :] < headdim))


@triton.jit
def _bwd_kernel_one_col_block(
    start_n,
    Q,
    K,
    V,
    Bias,
    DO,
    DQ,
    DK,
    DV,
    LSE,
    D,
    softmax_scale,
    stride_qm,
    stride_kn,
    stride_vn,
    stride_bm,
    stride_dom,
    stride_dqm,
    stride_dkn,
    stride_dvn,
    seqlen_q,
    seqlen_k,
    headdim,
    ATOMIC_ADD: tl.constexpr,
    BIAS_TYPE: tl.constexpr,
    IS_CAUSAL: tl.constexpr,
    BLOCK_HEADDIM: tl.constexpr,
    EVEN_M: tl.constexpr,
    EVEN_N: tl.constexpr,
    EVEN_HEADDIM: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # We need to make sure begin_m is a multiple of BLOCK_M (not BLOCK_N)
    begin_m = 0 if not IS_CAUSAL else ((start_n * BLOCK_N) // BLOCK_M) * BLOCK_M
    # initialize row/col offsets
    offs_qm = begin_m + tl.arange(0, BLOCK_M)
    offs_n = start_n * BLOCK_N + tl.arange(0, BLOCK_N)
    offs_m = tl.arange(0, BLOCK_M)
    offs_d = tl.arange(0, BLOCK_HEADDIM)
    # initialize pointers to value-like data
    q_ptrs = Q + (offs_qm[:, None] * stride_qm + offs_d[None, :])
    k_ptrs = K + (offs_n[:, None] * stride_kn + offs_d[None, :])
    v_ptrs = V + (offs_n[:, None] * stride_vn + offs_d[None, :])
    do_ptrs = DO + (offs_qm[:, None] * stride_dom + offs_d[None, :])
    dq_ptrs = DQ + (offs_qm[:, None] * stride_dqm + offs_d[None, :])
    if BIAS_TYPE == "vector":
        b_ptrs = Bias + offs_n
    elif BIAS_TYPE == "matrix":
        b_ptrs = Bias + (offs_qm[:, None] * stride_bm + offs_n[None, :])
    # initialize dv and dk
    dv = tl.zeros([BLOCK_N, BLOCK_HEADDIM], dtype=tl.float32)
    dk = tl.zeros([BLOCK_N, BLOCK_HEADDIM], dtype=tl.float32)
    # There seems to be some problem with Triton pipelining that makes results wrong for
    # headdim=64, seqlen=(113, 255), bias_type='matrix'. In this case the for loop
    # may have zero step, and pipelining with the bias matrix could cause the problem.
    # So we just exit early.
    if begin_m >= seqlen_q:
        dv_ptrs = DV + (offs_n[:, None] * stride_dvn + offs_d[None, :])
        dk_ptrs = DK + (offs_n[:, None] * stride_dkn + offs_d[None, :])
        _bwd_store_dk_dv(
            dk_ptrs,
            dv_ptrs,
            dk,
            dv,
            offs_n,
            offs_d,
            seqlen_k,
            headdim,
            EVEN_M=EVEN_M,
            EVEN_N=EVEN_N,
            EVEN_HEADDIM=EVEN_HEADDIM,
        )
        return
    # k and v stay in SRAM throughout
    # [2022-10-30] TD: Same bug as the fwd. In the case of EVEN_N=True and EVEN_M=False,
    # if we just call tl.load(k_ptrs), we get the wrong output!
    if EVEN_N & EVEN_M:
        if EVEN_HEADDIM:
            k = tl.load(k_ptrs)
            v = tl.load(v_ptrs)
        else:
            k = tl.load(k_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
            v = tl.load(v_ptrs, mask=offs_d[None, :] < headdim, other=0.0)
    else:
        if EVEN_HEADDIM:
            k = tl.load(k_ptrs, mask=offs_n[:, None] < seqlen_k, other=0.0)
            v = tl.load(v_ptrs, mask=offs_n[:, None] < seqlen_k, other=0.0)
        else:
            k = tl.load(k_ptrs, mask=(offs_n[:, None] < seqlen_k) & (offs_d[None, :] < headdim), other=0.0)
            v = tl.load(v_ptrs, mask=(offs_n[:, None] < seqlen_k) & (offs_d[None, :] < headdim), other=0.0)
    # loop over rows
    num_block_m = tl.cdiv(seqlen_q, BLOCK_M)
    for start_m in range(begin_m, num_block_m * BLOCK_M, BLOCK_M):
        start_m = tl.multiple_of(start_m, BLOCK_M)
        offs_m_curr = start_m + offs_m
        # load q, k, v, do on-chip
        # Same bug as below. Otherwise gives wrong result for headdim=40, seqlen=(128, 117)
        if EVEN_M & EVEN_HEADDIM:
            q = tl.load(q_ptrs)
        else:
            if EVEN_HEADDIM:
                q = tl.load(q_ptrs, mask=offs_m_curr[:, None] < seqlen_q, other=0.0)
            else:
                q = tl.load(
                    q_ptrs,
                    mask=(offs_m_curr[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
                    other=0.0,
                )
        # recompute p = softmax(qk, dim=-1).T
        qk = tl.dot(q, k, trans_b=True)
        # Trying to combine the two masks seem to make the result wrong
        if not EVEN_N:  # Need to mask out otherwise the softmax is wrong
            qk = tl.where(offs_n[None, :] < seqlen_k, qk, float("-inf"))
        if IS_CAUSAL:
            qk = tl.where(offs_m_curr[:, None] >= (offs_n[None, :]), qk, float("-inf"))
        if BIAS_TYPE != "none":
            tl.debug_barrier()  # Race condition otherwise
            if BIAS_TYPE == "vector":
                if EVEN_N:
                    bias = tl.load(b_ptrs).to(tl.float32)
                else:
                    bias = tl.load(b_ptrs, mask=offs_n < seqlen_k, other=0.0).to(tl.float32)
                bias = bias[None, :]
            elif BIAS_TYPE == "matrix":
                if EVEN_M & EVEN_N:
                    bias = tl.load(b_ptrs).to(tl.float32)
                else:
                    bias = tl.load(
                        b_ptrs,
                        mask=(offs_m_curr[:, None] < seqlen_q) & (offs_n[None, :] < seqlen_k),
                        other=0.0,
                    ).to(tl.float32)
            qk = qk * softmax_scale + bias
        # There seems to be a race condition when headdim=48/96, and dq, dk, dv are wrong.
        # Also wrong for headdim=64.
        if not (EVEN_M & EVEN_HEADDIM):
            tl.debug_barrier()
        lse_i = tl.load(LSE + offs_m_curr)
        if BIAS_TYPE == "none":
            p = tl.exp(qk * softmax_scale - lse_i[:, None])
        else:
            p = tl.exp(qk - lse_i[:, None])
        # compute dv
        # [2022-10-30] TD: A Triton bug: if EVEN_M=True and EVEN_HEADDIM=False, if we call
        # do = tl.load(do_ptrs, mask=offs_d[None, :] < headdim, other=0.0), we get wrong outputs
        # in the case of headdim=48/96, seqlen_q & seqlen_k >= 512. If headdim=40 or seqlen < 512,
        # the output is correct.
        if EVEN_M & EVEN_HEADDIM:
            do = tl.load(do_ptrs)
        else:
            # [2022-11-01] TD: Triton bug, there's a race condition if we just use m_mask and not d_mask.
            do = tl.load(
                do_ptrs,
                mask=(offs_m_curr[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
                other=0.0,
            )
        dv += tl.dot(p.to(do.dtype), do, trans_a=True)
        # compute dp = dot(v, do)
        # There seems to be a race condition when headdim=48/96, and dq, dk are wrong.
        # Also wrong for headdim=128, seqlen=(108, 256), and ATOMIC_ADD=True
        # Also wrong for headdim=64, seqlen=(1023, 1024), and ATOMIC_ADD=False
        if not (EVEN_M & EVEN_HEADDIM):
            tl.debug_barrier()
        dp = tl.dot(do, v, trans_b=True)
        # There's a race condition for headdim=48
        if not EVEN_HEADDIM:
            tl.debug_barrier()
        # compute ds = p * (dp - delta[:, None])
        # Putting the subtraction after the dp matmul (instead of before) is slightly faster
        Di = tl.load(D + offs_m_curr)
        # Converting ds to q.dtype here reduces register pressure and makes it much faster
        # for BLOCK_HEADDIM=128
        ds = (p * (dp - Di[:, None]) * softmax_scale).to(q.dtype)
        # compute dk = dot(ds.T, q)
        dk += tl.dot(ds, q, trans_a=True)
        # compute dq
        if not (EVEN_M & EVEN_HEADDIM):  # Otherewise there's a race condition when BIAS_TYPE='matrix'
            tl.debug_barrier()
        if not ATOMIC_ADD:
            if EVEN_M & EVEN_HEADDIM:  # Race condition if we just do EVEN_M
                dq = tl.load(dq_ptrs, eviction_policy="evict_last")
                dq += tl.dot(ds, k)
                tl.store(dq_ptrs, dq, eviction_policy="evict_last")
            else:
                if EVEN_HEADDIM:
                    dq = tl.load(
                        dq_ptrs,
                        mask=offs_m_curr[:, None] < seqlen_q,
                        other=0.0,
                        eviction_policy="evict_last",
                    )
                    dq += tl.dot(ds, k)
                    tl.store(
                        dq_ptrs,
                        dq,
                        mask=offs_m_curr[:, None] < seqlen_q,
                        eviction_policy="evict_last",
                    )
                else:
                    dq = tl.load(
                        dq_ptrs,
                        mask=(offs_m_curr[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
                        other=0.0,
                        eviction_policy="evict_last",
                    )
                    dq += tl.dot(ds, k)
                    tl.store(
                        dq_ptrs,
                        dq,
                        mask=(offs_m_curr[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
                        eviction_policy="evict_last",
                    )
        else:  # If we're parallelizing across the seqlen_k dimension
            dq = tl.dot(ds, k)
            if EVEN_M & EVEN_HEADDIM:  # Race condition if we just do EVEN_M
                tl.atomic_add(dq_ptrs, dq)
            else:
                if EVEN_HEADDIM:
                    tl.atomic_add(dq_ptrs, dq, mask=offs_m_curr[:, None] < seqlen_q)
                else:
                    tl.atomic_add(
                        dq_ptrs,
                        dq,
                        mask=(offs_m_curr[:, None] < seqlen_q) & (offs_d[None, :] < headdim),
                    )
        # increment pointers
        dq_ptrs += BLOCK_M * stride_dqm
        q_ptrs += BLOCK_M * stride_qm
        do_ptrs += BLOCK_M * stride_dom
        if BIAS_TYPE == "matrix":
            b_ptrs += BLOCK_M * stride_bm
    # write-back
    dv_ptrs = DV + (offs_n[:, None] * stride_dvn + offs_d[None, :])
    dk_ptrs = DK + (offs_n[:, None] * stride_dkn + offs_d[None, :])
    _bwd_store_dk_dv(
        dk_ptrs,
        dv_ptrs,
        dk,
        dv,
        offs_n,
        offs_d,
        seqlen_k,
        headdim,
        EVEN_M=EVEN_M,
        EVEN_N=EVEN_N,
        EVEN_HEADDIM=EVEN_HEADDIM,
    )


def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()


@triton.autotune(
    configs=[
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_N": 128, "SEQUENCE_PARALLEL": False},
            num_warps=8,
            num_stages=1,
            pre_hook=init_to_zero("DQ"),
        ),
        triton.Config(
            {"BLOCK_M": 128, "BLOCK_N": 128, "SEQUENCE_PARALLEL": True},
            num_warps=8,
            num_stages=1,
            pre_hook=init_to_zero("DQ"),
        ),
        # Other configs seem to give wrong results when seqlen_q % 128 != 0, disabling them for now
        # # Kernel is buggy (give wrong result) if we set BLOCK_m=128, BLOCK_n=64, num_warps=*4*
        # triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "SEQUENCE_PARALLEL": False}, num_warps=8, num_stages=1, pre_hook=init_to_zero('DQ')),
        # triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "SEQUENCE_PARALLEL": True}, num_warps=8, num_stages=1, pre_hook=init_to_zero('DQ')),
        # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64, "SEQUENCE_PARALLEL": False}, num_warps=4, num_stages=1, pre_hook=init_to_zero('DQ')),
        # triton.Config({"BLOCK_M": 64, "BLOCK_N": 64, "SEQUENCE_PARALLEL": True}, num_warps=4, num_stages=1, pre_hook=init_to_zero('DQ')),
    ],
    key=["CACHE_KEY_SEQLEN_Q", "CACHE_KEY_SEQLEN_K", "BIAS_TYPE", "IS_CAUSAL", "BLOCK_HEADDIM"],
)
@triton.heuristics(
    {
        "EVEN_M": lambda args: args["seqlen_q"] % args["BLOCK_M"] == 0,
        "EVEN_N": lambda args: args["seqlen_k"] % args["BLOCK_N"] == 0,
        "EVEN_HEADDIM": lambda args: args["headdim"] == args["BLOCK_HEADDIM"],
    }
)
@triton.jit
def _bwd_kernel(
    Q,
    K,
    V,
    Bias,
    DO,
    DQ,
    DK,
    DV,
    LSE,
    D,
    softmax_scale,
    stride_qb,
    stride_qh,
    stride_qm,
    stride_kb,
    stride_kh,
    stride_kn,
    stride_vb,
    stride_vh,
    stride_vn,
    stride_bb,
    stride_bh,
    stride_bm,
    stride_dob,
    stride_doh,
    stride_dom,
    stride_dqb,
    stride_dqh,
    stride_dqm,
    stride_dkb,
    stride_dkh,
    stride_dkn,
    stride_dvb,
    stride_dvh,
    stride_dvn,
    nheads,
    seqlen_q,
    seqlen_k,
    seqlen_q_rounded,
    headdim,
    CACHE_KEY_SEQLEN_Q,
    CACHE_KEY_SEQLEN_K,
    BIAS_TYPE: tl.constexpr,
    IS_CAUSAL: tl.constexpr,
    BLOCK_HEADDIM: tl.constexpr,
    SEQUENCE_PARALLEL: tl.constexpr,
    EVEN_M: tl.constexpr,
    EVEN_N: tl.constexpr,
    EVEN_HEADDIM: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    off_hb = tl.program_id(1)
    off_b = off_hb // nheads
    off_h = off_hb % nheads
    # offset pointers for batch/head
    Q += off_b * stride_qb + off_h * stride_qh
    K += off_b * stride_kb + off_h * stride_kh
    V += off_b * stride_vb + off_h * stride_vh
    DO += off_b * stride_dob + off_h * stride_doh
    DQ += off_b * stride_dqb + off_h * stride_dqh
    DK += off_b * stride_dkb + off_h * stride_dkh
    DV += off_b * stride_dvb + off_h * stride_dvh
    if BIAS_TYPE != "none":
        Bias += off_b * stride_bb + off_h * stride_bh
    # pointer to row-wise quantities in value-like data
    D += off_hb * seqlen_q_rounded
    LSE += off_hb * seqlen_q_rounded
    if not SEQUENCE_PARALLEL:
        num_block_n = tl.cdiv(seqlen_k, BLOCK_N)
        for start_n in range(num_block_n):
            _bwd_kernel_one_col_block(
                start_n,
                Q,
                K,
                V,
                Bias,
                DO,
                DQ,
                DK,
                DV,
                LSE,
                D,
                softmax_scale,
                stride_qm,
                stride_kn,
                stride_vn,
                stride_bm,
                stride_dom,
                stride_dqm,
                stride_dkn,
                stride_dvn,
                seqlen_q,
                seqlen_k,
                headdim,
                ATOMIC_ADD=False,
                BIAS_TYPE=BIAS_TYPE,
                IS_CAUSAL=IS_CAUSAL,
                BLOCK_HEADDIM=BLOCK_HEADDIM,
                EVEN_M=EVEN_M,
                EVEN_N=EVEN_N,
                EVEN_HEADDIM=EVEN_HEADDIM,
                BLOCK_M=BLOCK_M,
                BLOCK_N=BLOCK_N,
            )
    else:
        start_n = tl.program_id(0)
        _bwd_kernel_one_col_block(
            start_n,
            Q,
            K,
            V,
            Bias,
            DO,
            DQ,
            DK,
            DV,
            LSE,
            D,
            softmax_scale,
            stride_qm,
            stride_kn,
            stride_vn,
            stride_bm,
            stride_dom,
            stride_dqm,
            stride_dkn,
            stride_dvn,
            seqlen_q,
            seqlen_k,
            headdim,
            ATOMIC_ADD=True,
            BIAS_TYPE=BIAS_TYPE,
            IS_CAUSAL=IS_CAUSAL,
            BLOCK_HEADDIM=BLOCK_HEADDIM,
            EVEN_M=EVEN_M,
            EVEN_N=EVEN_N,
            EVEN_HEADDIM=EVEN_HEADDIM,
            BLOCK_M=BLOCK_M,
            BLOCK_N=BLOCK_N,
        )


def flash_attn_forward(q, k, v, bias=None, **kwargs):
    # shape constraints
    batch, seqlen_q, nheads, d = q.shape
    _, seqlen_k, _, _ = k.shape
    assert k.shape == (batch, seqlen_k, nheads, d)
    assert v.shape == (batch, seqlen_k, nheads, d)
    assert d <= 128, "FlashAttention only support head dimensions up to 128"
    assert q.dtype == k.dtype == v.dtype, "All tensors must have the same type"
    assert q.dtype in [torch.float16, torch.bfloat16], "Only support fp16 and bf16"
    assert q.is_cuda and k.is_cuda and v.is_cuda

    causal = kwargs.get("causal", 0) == 1
    softmax_scale = kwargs.get("softmax_scale", 1.0 / math.sqrt(d))
    has_bias = bias is not None
    bias_type = "none"
    if has_bias:
        assert bias.dtype in [q.dtype, torch.float]
        assert bias.is_cuda
        assert bias.dim() == 4
        if bias.stride(-1) != 1:
            bias = bias.contiguous()
        if bias.shape[2:] == (1, seqlen_k):
            bias_type = "vector"
        elif bias.shape[2:] == (seqlen_q, seqlen_k):
            bias_type = "matrix"
        else:
            raise RuntimeError("Last 2 dimensions of bias must be (1, seqlen_k) or (seqlen_q, seqlen_k)")
        bias = bias.expand(batch, nheads, seqlen_q, seqlen_k)
    bias_strides = (bias.stride(0), bias.stride(1), bias.stride(2)) if has_bias else (0, 0, 0)

    seqlen_q_rounded = math.ceil(seqlen_q / 128) * 128
    lse = torch.empty((batch, nheads, seqlen_q_rounded), device=q.device, dtype=torch.float32)
    tmp = torch.empty((batch, nheads, seqlen_q_rounded), device=q.device, dtype=torch.float32)
    o = torch.empty_like(q)

    BLOCK_HEADDIM = max(triton.next_power_of_2(d), 16)
    BLOCK = 128
    num_warps = 4 if d <= 64 else 8
    grid = lambda META: (triton.cdiv(seqlen_q, META["BLOCK_M"]), batch * nheads)
    _fwd_kernel[grid](
        q,
        k,
        v,
        bias,
        o,
        lse,
        tmp,
        softmax_scale,
        q.stride(0),
        q.stride(2),
        q.stride(1),
        k.stride(0),
        k.stride(2),
        k.stride(1),
        v.stride(0),
        v.stride(2),
        v.stride(1),
        *bias_strides,
        o.stride(0),
        o.stride(2),
        o.stride(1),
        nheads,
        seqlen_q,
        seqlen_k,
        seqlen_q_rounded,
        d,
        seqlen_q // 32,
        seqlen_k // 32,  # key for triton cache (limit number of compilations)
        # Can't use kwargs here because triton autotune expects key to be args, not kwargs
        # IS_CAUSAL=causal, BLOCK_HEADDIM=d,
        bias_type,
        causal,
        BLOCK_HEADDIM,
        BLOCK_M=BLOCK,
        BLOCK_N=BLOCK,
        num_warps=num_warps,
        num_stages=1,
    )
    return o, lse


def flash_attn_backward(do, q, k, v, o, lse, bias=None, **kwargs):
    dq = torch.empty_like(q)
    dk = torch.empty_like(k)
    dv = torch.empty_like(v)
    # Make sure that the last dimension is contiguous
    if do.stride(-1) != 1:
        do = do.contiguous()
    batch, seqlen_q, nheads, d = q.shape
    _, seqlen_k, _, _ = k.shape
    # assert d in {16, 32, 64, 128}
    assert d <= 128
    seqlen_q_rounded = math.ceil(seqlen_q / 128) * 128
    assert lse.shape == (batch, nheads, seqlen_q_rounded)
    assert q.stride(-1) == k.stride(-1) == v.stride(-1) == o.stride(-1) == 1
    assert dq.stride(-1) == dk.stride(-1) == dv.stride(-1) == 1

    causal = kwargs.get("causal", 0) == 1
    softmax_scale = kwargs.get("softmax_scale", 1.0 / math.sqrt(d))
    # dq_accum = torch.zeros_like(q, dtype=torch.float32)
    dq_accum = torch.empty_like(q, dtype=torch.float32)
    delta = torch.empty_like(lse)
    # delta = torch.zeros_like(lse)

    BLOCK_HEADDIM = max(triton.next_power_of_2(d), 16)
    grid = lambda META: (triton.cdiv(seqlen_q, META["BLOCK_M"]), batch * nheads)
    _bwd_preprocess_do_o_dot[grid](
        o,
        do,
        delta,
        o.stride(0),
        o.stride(2),
        o.stride(1),
        do.stride(0),
        do.stride(2),
        do.stride(1),
        nheads,
        seqlen_q,
        seqlen_q_rounded,
        d,
        BLOCK_M=128,
        BLOCK_HEADDIM=BLOCK_HEADDIM,
    )

    has_bias = bias is not None
    bias_type = "none"
    if has_bias:
        assert bias.dtype in [q.dtype, torch.float]
        assert bias.is_cuda
        assert bias.dim() == 4
        assert bias.stride(-1) == 1
        if bias.shape[2:] == (1, seqlen_k):
            bias_type = "vector"
        elif bias.shape[2:] == (seqlen_q, seqlen_k):
            bias_type = "matrix"
        else:
            raise RuntimeError("Last 2 dimensions of bias must be (1, seqlen_k) or (seqlen_q, seqlen_k)")
        bias = bias.expand(batch, nheads, seqlen_q, seqlen_k)
    bias_strides = (bias.stride(0), bias.stride(1), bias.stride(2)) if has_bias else (0, 0, 0)

    # BLOCK_M = 128
    # BLOCK_N = 64
    # num_warps = 4
    grid = lambda META: (
        triton.cdiv(seqlen_k, META["BLOCK_N"]) if META["SEQUENCE_PARALLEL"] else 1,
        batch * nheads,
    )
    _bwd_kernel[grid](
        q,
        k,
        v,
        bias,
        do,
        dq_accum,
        dk,
        dv,
        lse,
        delta,
        softmax_scale,
        q.stride(0),
        q.stride(2),
        q.stride(1),
        k.stride(0),
        k.stride(2),
        k.stride(1),
        v.stride(0),
        v.stride(2),
        v.stride(1),
        *bias_strides,
        do.stride(0),
        do.stride(2),
        do.stride(1),
        dq_accum.stride(0),
        dq_accum.stride(2),
        dq_accum.stride(1),
        dk.stride(0),
        dk.stride(2),
        dk.stride(1),
        dv.stride(0),
        dv.stride(2),
        dv.stride(1),
        nheads,
        seqlen_q,
        seqlen_k,
        seqlen_q_rounded,
        d,
        seqlen_q // 32,
        seqlen_k // 32,  # key for triton cache (limit number of compilations)
        # Can't use kwargs here because triton autotune expects key to be args, not kwargs
        # IS_CAUSAL=causal, BLOCK_HEADDIM=d,
        bias_type,
        causal,
        BLOCK_HEADDIM,
        # SEQUENCE_PARALLEL=False,
        # BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N,
        # num_warps=num_warps,
        # num_stages=1,
    )
    dq.copy_(dq_accum)
    return dq, dk, dv


def _make_flash_attention_nodes(
    idx: int,
    q: str,
    k: str,
    v: str,
    y: str,
    dy: str,
    dq: str,
    dk: str,
    dv: str,
    bias: str,
    scale: float,
):
    logsumexp = helper.make_tensor_value_info("logsumexp_" + str(idx), TensorProto.FLOAT, [])
    fwd_node = helper.make_node(
        "TritonOp",
        [q, k, v, bias],
        [y, logsumexp.name],
        "TritonOp_Flash_Attn_Fwd_" + str(idx),
        None,
        "com.microsoft",
        func_name="flash_attn_forward",
        causal=0,
        softmax_scale=scale,
    )
    bwd_node = helper.make_node(
        "TritonOp",
        [dy, q, k, v, y, logsumexp.name, bias],
        [dq, dk, dv],
        "TritonOp_Flash_Attn_Bwd_" + str(idx),
        None,
        "com.microsoft",
        func_name="flash_attn_backward",
        causal=0,
        softmax_scale=scale,
    )
    return [fwd_node, bwd_node], [logsumexp]


# Without causal mask, without Dropout. For example, BERT model in HuggingFace.
_PATTERN_0: list[tuple[str, bool, list[tuple[int, int, int]]]] = [
    ("MatMul", False, []),  # 0
    ("Transpose", True, [(0, 0, 0)]),  # 1
    ("Transpose", True, [(0, 0, 1)]),  # 2
    ("Div", False, [(0, 0, 0)]),  # 3
    ("Add", False, [(3, 0, 0)]),  # 4
    ("Softmax", False, [(4, 0, 0)]),  # 5
    ("MatMul", False, [(5, 0, 0)]),  # 6
    ("Transpose", True, [(6, 0, 1)]),  # 7
    ("Transpose", False, [(6, 0, 0)]),  # 8
    ("FusedMatMul", False, [(7, 0, 1)]),  # 9
    ("SoftmaxGrad_13", False, [(9, 0, 0), (5, 0, 1)]),  # 10
    ("Identity", False, [(10, 0, 0)]),  # 11
    ("Div", False, [(11, 0, 0)]),  # 12
    ("Identity", False, [(12, 0, 0)]),  # 13
    ("FusedMatMul", False, [(2, 0, 1), (13, 0, 0)]),  # 14
    ("FusedMatMul", False, [(1, 0, 0), (13, 0, 1)]),  # 15
    ("FusedMatMul", False, [(5, 0, 0)]),  # 16
    ("Transpose", True, [(16, 0, 1)]),  # 17
    ("Transpose", False, [(14, 0, 0)]),  # 18
    ("Transpose", False, [(15, 0, 0)]),  # 19
    ("Transpose", False, [(16, 0, 0)]),  # 20
]


def _optimize_for_pattern_0(matcher: GraphMatcher, idx: int, nodes: list[NodeProto]):
    # Check forward only as the backward is expected to be consistent if it's built correctly.
    scale_value = matcher.get_constant_value(nodes[3].input[1])
    if not (
        check_attribute_value(nodes[1], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[2], "perm", [0, 2, 3, 1])
        and scale_value is not None
        and check_attribute_value(nodes[7], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[8], "perm", [0, 2, 1, 3])
    ):
        return [], [], []

    nodes_to_add, new_value_infos = _make_flash_attention_nodes(
        idx,
        nodes[1].input[0],
        nodes[2].input[0],
        nodes[7].input[0],
        nodes[8].output[0],
        nodes[17].input[0],
        nodes[18].output[0],
        nodes[19].output[0],
        nodes[20].output[0],
        nodes[4].input[1],
        1 / float(scale_value[0] if isinstance(scale_value, list) else scale_value),
    )
    return nodes, nodes_to_add, new_value_infos


# llama2+peft, k doesn't require grad.
_PATTERN_1: list[tuple[str, bool, list[tuple[int, int, int]]]] = [
    ("MatMul", False, []),  # 0
    ("Transpose", True, [(0, 0, 1)]),  # 1
    ("Div", False, [(0, 0, 0)]),  # 2
    ("Add", False, [(2, 0, 0)]),  # 3
    ("Softmax", False, [(3, 0, 0)]),  # 4
    ("MatMul", False, [(4, 0, 0)]),  # 5
    ("Transpose", True, [(5, 0, 1)]),  # 6
    ("Identity", False, [(6, 0, 0)]),  # 7
    ("YieldOp", False, [(7, 0, -1)]),  # 8
    ("Transpose", False, [(5, 0, 0)]),  # 9
    ("FusedMatMul", False, [(6, 0, 1)]),  # 10
    ("SoftmaxGrad_13", False, [(10, 0, 0), (4, 0, 1)]),  # 11
    ("Identity", False, [(11, 0, 0)]),  # 12
    ("Div", False, [(12, 0, 0)]),  # 13
    ("Identity", False, [(13, 0, 0)]),  # 14
    ("FusedMatMul", False, [(1, 0, 1), (14, 0, 0)]),  # 15
    ("FusedMatMul", False, [(4, 0, 0)]),  # 16
    ("Transpose", True, [(16, 0, 1)]),  # 17
    ("Sum", False, [(16, 0, 0)]),  # 18
    ("Transpose", False, [(18, 0, 0)]),  # 19
]


def _optimize_for_pattern_1(matcher: GraphProto, idx: int, nodes: list[NodeProto]):
    # Check forward only as the backward is expected to be consistent if it's built correctly.
    scale_value = matcher.get_constant_value(nodes[2].input[1])
    if not (
        check_attribute_value(nodes[1], "perm", [0, 1, 3, 2])
        and scale_value is not None
        and check_attribute_value(nodes[6], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[9], "perm", [0, 2, 1, 3])
        and matcher.get_consumer_count(nodes[14].output[0]) == 1
    ):
        return [], [], []

    dtype, _ = matcher.get_type_and_shape(nodes[0].input[0])
    assert dtype is not None
    trans_q_tensor = helper.make_tensor_value_info("trans_q_" + str(idx), dtype, None)
    trans_q_grad_tensor = helper.make_tensor_value_info("trans_q_grad_" + str(idx), dtype, None)
    trans_k_tensor = helper.make_tensor_value_info("trans_k_" + str(idx), dtype, None)
    trans_q = helper.make_node(
        "Transpose", [nodes[0].input[0]], [trans_q_tensor.name], "Trans_Q_" + str(idx), perm=[0, 2, 1, 3]
    )
    trans_q_grad = helper.make_node(
        "Transpose", [trans_q_grad_tensor.name], [nodes[15].output[0]], "Trans_Q_Grad_" + str(idx), perm=[0, 2, 1, 3]
    )
    trans_k = helper.make_node(
        "Transpose", [nodes[1].input[0]], [trans_k_tensor.name], "Trans_K_" + str(idx), perm=[0, 2, 1, 3]
    )
    nodes[19].input[0] = nodes[18].input[1]
    v_grad = nodes[19].output[0]
    nodes[19].output[0] = nodes[18].output[0]
    nodes[18].input[1] = nodes[18].output[0]
    nodes[18].output[0] = v_grad
    nodes_to_add, new_value_infos = _make_flash_attention_nodes(
        idx,
        trans_q_tensor.name,
        trans_k_tensor.name,
        nodes[6].input[0],
        nodes[9].output[0],
        nodes[17].input[0],
        trans_q_grad_tensor.name,
        "",
        nodes[16].output[0],
        nodes[3].input[1],
        1 / float(scale_value[0] if isinstance(scale_value, list) else scale_value),
    )
    nodes_to_remove = nodes[:6] + nodes[9:18]
    nodes_to_add.extend([trans_q, trans_q_grad, trans_k])
    new_value_infos.extend([trans_q_tensor, trans_q_grad_tensor, trans_k_tensor])
    return nodes_to_remove, nodes_to_add, new_value_infos


# llama2+peft, k requires grad.
_PATTERN_2: list[tuple[str, bool, list[tuple[int, int, int]]]] = [
    ("MatMul", False, []),  # 0
    ("Transpose", True, [(0, 0, 1)]),  # 1
    ("Div", False, [(0, 0, 0)]),  # 2
    ("Add", False, [(2, 0, 0)]),  # 3
    ("Softmax", False, [(3, 0, 0)]),  # 4
    ("MatMul", False, [(4, 0, 0)]),  # 5
    ("Transpose", True, [(5, 0, 1)]),  # 6
    ("Identity", False, [(6, 0, 0)]),  # 7
    ("YieldOp", False, [(7, 0, -1)]),  # 8
    ("Transpose", False, [(5, 0, 0)]),  # 9
    ("FusedMatMul", False, [(6, 0, 1)]),  # 10
    ("SoftmaxGrad_13", False, [(10, 0, 0), (4, 0, 1)]),  # 11
    ("Identity", False, [(11, 0, 0)]),  # 12
    ("Div", False, [(12, 0, 0)]),  # 13
    ("Identity", False, [(13, 0, 0)]),  # 14
    ("FusedMatMul", False, [(1, 0, 1), (14, 0, 0)]),  # 15
    ("FusedMatMul", False, [(14, 0, 1)]),  # 16
    ("Transpose", False, [(16, 0, 0)]),  # 17
    ("FusedMatMul", False, [(4, 0, 0)]),  # 18
    ("Transpose", True, [(18, 0, 1)]),  # 19
    ("Sum", False, [(18, 0, 0)]),  # 20
    ("Transpose", False, [(20, 0, 0)]),  # 21
]


def _aptimize_for_pattern_2(matcher: GraphProto, idx: int, nodes: list[NodeProto]):
    # Check forward only as the backward is expected to be consistent if it's built correctly.
    scale_value = matcher.get_constant_value(nodes[2].input[1])
    if not (
        check_attribute_value(nodes[1], "perm", [0, 1, 3, 2])
        and scale_value is not None
        and check_attribute_value(nodes[6], "perm", [0, 2, 1, 3])
        and check_attribute_value(nodes[9], "perm", [0, 2, 1, 3])
        and matcher.get_consumer_count(nodes[14].output[0]) == 2
    ):
        return [], [], []

    dtype, _ = matcher.get_type_and_shape(nodes[0].input[0])
    assert dtype is not None
    trans_q_tensor = helper.make_tensor_value_info("trans_q_" + str(idx), dtype, None)
    trans_q_grad_tensor = helper.make_tensor_value_info("trans_q_grad_" + str(idx), dtype, None)
    trans_k_tensor = helper.make_tensor_value_info("trans_k_" + str(idx), dtype, None)
    trans_k_grad_tensor = helper.make_tensor_value_info("trans_k_grad_" + str(idx), dtype, None)
    trans_q = helper.make_node(
        "Transpose", [nodes[0].input[0]], [trans_q_tensor.name], "Trans_Q_" + str(idx), perm=[0, 2, 1, 3]
    )
    trans_q_grad = helper.make_node(
        "Transpose", [trans_q_grad_tensor.name], [nodes[15].output[0]], "Trans_Q_Grad_" + str(idx), perm=[0, 2, 1, 3]
    )
    trans_k = helper.make_node(
        "Transpose", [nodes[1].input[0]], [trans_k_tensor.name], "Trans_K_" + str(idx), perm=[0, 2, 1, 3]
    )
    trans_k_grad = helper.make_node(
        "Transpose", [trans_k_grad_tensor.name], [nodes[17].output[0]], "Trans_K_Grad_" + str(idx), perm=[0, 2, 1, 3]
    )
    nodes[21].input[0] = nodes[20].input[1]
    v_grad = nodes[21].output[0]
    nodes[21].output[0] = nodes[20].output[0]
    nodes[20].input[1] = nodes[20].output[0]
    nodes[20].output[0] = v_grad
    nodes_to_add, new_value_infos = _make_flash_attention_nodes(
        idx,
        trans_q_tensor.name,
        trans_k_tensor.name,
        nodes[6].input[0],
        nodes[9].output[0],
        nodes[19].input[0],
        trans_q_grad_tensor.name,
        trans_k_grad_tensor.name,
        nodes[18].output[0],
        nodes[3].input[1],
        1 / float(scale_value[0] if isinstance(scale_value, list) else scale_value),
    )
    nodes_to_remove = nodes[:6] + nodes[9:20]
    nodes_to_add.extend([trans_q, trans_q_grad, trans_k, trans_k_grad])
    new_value_infos.extend([trans_q_tensor, trans_q_grad_tensor, trans_k_tensor, trans_k_grad_tensor])
    return nodes_to_remove, nodes_to_add, new_value_infos


# TODO: add pattern to support attention with causal mask, such as GPT2 in HuggingFace.
_PATTERNS = [
    (_PATTERN_0, _optimize_for_pattern_0),
    (_PATTERN_1, _optimize_for_pattern_1),
    (_PATTERN_2, _aptimize_for_pattern_2),
]


@register_graph_optimizer(devices="cuda")
def optimize_graph_for_flash_attention(graph: GraphProto):
    nodes_to_remove = []
    nodes_to_add = []
    new_value_infos = []
    matcher = GraphMatcher(graph)
    idx = 0
    for pattern_tuple in _PATTERNS:
        for nodes in matcher.match_pattern(pattern_tuple[0]):
            remove_nodes, add_nodes, add_value_infos = pattern_tuple[1](matcher, idx, nodes)
            if len(add_nodes) > 0:
                nodes_to_remove.extend(remove_nodes)
                nodes_to_add.extend(add_nodes)
                new_value_infos.extend(add_value_infos)
                idx += 1
    update_graph(graph, nodes_to_remove, nodes_to_add, new_value_infos)
